/* map.c  */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/* map lines module */



#include <stdio.h>
#ifndef SEEK_SET
#  define SEEK_SET 0
#endif

#include "globals.h"
#include "graphics.h"
#include "topo.h"



#define MAXVERT 30000
#define MAXSEG  1000

static float vert[MAXVERT][3];
static int start[MAXSEG], len[MAXSEG];

static float xscale, yscale, zscale;
static float transx[MAXTIMES], transy[MAXTIMES];
static int segcount, vertcount;



/*** init_map *********************************************************
   Read a map file.
   Input:  mapname - name of map file
   Return:  0 = error, non-zero = success
**********************************************************************/
int init_map( mapname )
char *mapname;
{
   FILE *mapfile;
   int nsegs;
   int *dir, *llval, *verts;
   int i, j, offset, length, it;
   float lat, lon, el;

   mapfile = fopen( mapname, "r" );
   if (!mapfile)
      return 0;

   /* read number of segments (number of polylines) */
   if (fread( &nsegs, sizeof(int), 1, mapfile)!=1) {
      fclose( mapfile );
      return 0;
   }

   /* allocate and read segment directory */
   dir = (int *) malloc( nsegs * 6 * sizeof(int) );
   if (!dir) {
      fclose( mapfile );
      return 0;
   }
   if (fread( dir, sizeof(int), nsegs*6, mapfile)!=nsegs*6) {
      free( dir );
      fclose( mapfile );
      return 0;
   }

   xscale = (Xmax-Xmin) / (EastLon[0]-WestLon[0]);
   yscale = (Ymax-Ymin) / (NorthLat[0]-SouthLat[0]);
   zscale = (Zmax-Zmin) / (TopHgt[0]-BottomHgt[0]);

   segcount = vertcount = 0;
   start[0] = 0;
   len[0] = 0;

   for (i=0;i<nsegs && segcount<MAXSEG;i++) {
      /* get list of vertices for ith segment */
      offset = dir[i*6+4] * 4;
      if (fseek(mapfile,offset,SEEK_SET)) {
	 free( dir );
	 fclose( mapfile);
	 return 0;
      }
      length = dir[i*6+5];
      verts = (int *) malloc( length * sizeof(int) );
      if (fread( verts, sizeof(int), length, mapfile)!=length) {
	 free( dir );
	 free( verts );
	 fclose( mapfile );
	 return 0;
      }

      for (j=0;j<length/2 && vertcount<MAXVERT;j++) {
	 lat = (float) verts[j*2] * 0.0001;
	 lon = (float) verts[j*2+1] * 0.0001;
	 el = elevation(lat,lon,NULL) / 1000.0;

	 add_vertex( lon, lat, el );
/*
	 vert[vertcount][0] = Xmin + lon * xscale;
	 vert[vertcount][1] = Ymin + lat * yscale;
	 vert[vertcount][2] = Zmax - (el-BottomHgt[0])*zscale - 0.01;
	 vertcount++;
*/
      }
      end_segment();

      free( verts );
   }

   free( dir );
   fclose( mapfile );

   /* calculate translation values for each time step */
   for (it=0;it<NumTimes;it++) {
      transx[it] = -WestLon[it] * xscale;
      transy[it] = -SouthLat[it] * yscale;
/*      printf("trans[%d] = %f %f\n", it, transx[it], transy[it] );*/
   }

   /*printf("done (segcount=%d, vertcount=%d)\n", segcount, vertcount);*/
   return 1;
}



static end_segment()
{
   segcount++;
   /* initialize next */
   start[segcount] = vertcount;
   len[segcount] = 0;
}



/* Add a new map line vertex.
   Input:  x - the longitude in degrees
           y - the latitude in degrees
           z - the elevation in meters
*/
static add_vertex( lon, lat, hgt )
float lon, lat, hgt;
{

   if (WestBound>=lon && lon>=EastBound
       && NorthBound>=lat && lat>=SouthBound) {
      /* add the vertex */
      vert[vertcount][0] = Xmin + lon * xscale;
      vert[vertcount][1] = Ymin + lat * yscale;
      vert[vertcount][2] = Zmax - (hgt-BottomBound)*zscale - 0.01;
      vertcount++;
      len[segcount]++;
   }
   else {
      /* vertex is outside of domain */
      if (len[segcount]>0) {
	 /* end the segment */
	 end_segment();
      }
      else {
	 /* segment wasn't started, do nothing */
      }
   }

}




/*** draw_map *********************************************************
   Draw the map.
   Input:  it - time step
           c - color
	   flat - 1 = draw flat map
	          0 = draw raised map
**********************************************************************/
void draw_map( it, c, flat )
int it, c, flat;
{
   float v[1000][3];
   int i, j, num, st;

   /* set line color and clip line segments to box bounds */
   set_color( c );
   if (flat)
      set_depthcue( 1 );
   if (!StaticDomain)
      set_box_clipping( 1, Xmin, Xmax, Ymin, Ymax );

   if (flat) {
      /* draw a flat map */
      for (i=0;i<segcount;i++) {
	 if (len[i]>1000)
	   num = 1000;
	 else
	   num = len[i];
	 st = start[i];
	 for (j=0;j<num;j++) {  /* fully vectorized */
	    v[j][0] = vert[st+j][0] + transx[it];
	    v[j][1] = vert[st+j][1] + transy[it];
	    v[j][2] = Zmax;
	 }
	 polyline( v, num );
      }
   }
   else {
      /* draw a map with heights */
      for (i=0;i<segcount;i++) {
	 if (len[i]>1000)
	   num = 1000;
	 else
	   num = len[i];
	 st = start[i];
	 for (j=0;j<num;j++) {  /* fully vectorized */
	    v[j][0] = vert[st+j][0] + transx[it];
	    v[j][1] = vert[st+j][1] + transy[it];
	    v[j][2] = vert[st+j][2];
	 }
	 polyline( v, num );
      }
   }

   if (!StaticDomain)
      set_box_clipping( 0, 0.0, 0.0, 0.0, 0.0 );
   set_depthcue( 0 );

}


